/*
 * Copyright (c) 2017 - 2020, GÉANT
 *
 * Licensed under the Apache License, Version 2.0 (the “License”); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an “AS IS” BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.geant.idpextension.oidc.attribute.transcoding.impl;

import java.util.Collections;
import java.util.List;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;

import org.geant.idpextension.oidc.attribute.transcoding.AbstractOIDCAttributeTranscoder;
import org.opensaml.profile.context.ProfileRequestContext;

import net.minidev.json.JSONArray;
import net.shibboleth.idp.attribute.AttributeDecodingException;
import net.shibboleth.idp.attribute.AttributeEncodingException;
import net.shibboleth.idp.attribute.ByteAttributeValue;
import net.shibboleth.idp.attribute.IdPAttribute;
import net.shibboleth.idp.attribute.IdPAttributeValue;
import net.shibboleth.idp.attribute.transcoding.TranscodingRule;
import net.shibboleth.utilities.java.support.annotation.constraint.NonnullElements;
import net.shibboleth.utilities.java.support.codec.Base64Support;

/**
 * Class encoding byte attributes to base64 encoded string json object. Name of the attribute will be set as the key.
 * The string contains base64 coded attribute value. If there are several attribute values they are delimited with
 * space. The output may be set also to array. The output may also be set to be int instead of b64. In that case each
 * value (byte[]) is converted to int array and placed into array.
 */
public class OIDCByteAttributeTranscoder extends AbstractOIDCAttributeTranscoder {

// Checkstyle: CyclomaticComplexity OFF
    /** {@inheritDoc} */
    @Override
    @Nullable protected Object encodeValues(@Nullable final ProfileRequestContext profileRequestContext,
            @Nonnull final IdPAttribute attribute, @Nonnull final TranscodingRule rule)
                    throws AttributeEncodingException {
        
        final Boolean asInteger = rule.getOrDefault(PROP_ASINTEGER, Boolean.class, false);
        final Boolean asArray = rule.getOrDefault(PROP_ASARRAY, Boolean.class, false);
        final String delimiter = rule.getOrDefault(PROP_STRING_DELIMITER, String.class, ""); 
        
        String attributeString = "";
        final JSONArray array = new JSONArray();
        for (final IdPAttributeValue value : attribute.getValues()) {
            if (value instanceof ByteAttributeValue && ((ByteAttributeValue) value).getValue() != null) {
                if (asInteger) {
                    final JSONArray innerArray = new JSONArray();
                    for (final byte byteValue : ((ByteAttributeValue) value).getValue()) {
                        innerArray.add((int) byteValue);
                    }
                    // each byte array is converted to JSON int array and placed into JSON array.
                    array.add(innerArray);
                } else {
                    if (attributeString.length() > 0) {
                        attributeString += delimiter;
                    }
                    attributeString +=
                            Base64Support.encode(((ByteAttributeValue) value).getValue(), Base64Support.UNCHUNKED);
                    if (asArray) {
                        array.add(attributeString.toString());
                        attributeString = "";
                    }
                }
            }
        }
        if (asArray || asInteger) {
            return array.size() == 0 ? null : array;
        } else {
            return attributeString.toString().isEmpty() ? null : attributeString.toString();
        }
    }
// Checkstyle: CyclomaticComplexity ON
    
    /** {@inheritDoc} */
    @Override
    @Nonnull @NonnullElements protected List<IdPAttributeValue> decodeValues(
            @Nullable final ProfileRequestContext profileRequestContext, @Nonnull final Object input,
            @Nonnull final TranscodingRule rule) throws AttributeDecodingException {
        
        // TODO: implement value decoding
        return Collections.emptyList();
    }

}