/*
 * Copyright (c) 2017 - 2020, GÉANT
 *
 * Licensed under the Apache License, Version 2.0 (the “License”); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an “AS IS” BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.geant.idpextension.oidc.attribute.transcoding.impl;

import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;

import org.geant.idpextension.oidc.attribute.transcoding.AbstractOIDCAttributeTranscoder;
import org.opensaml.core.xml.schema.XSBooleanValue;
import org.opensaml.profile.context.ProfileRequestContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.common.base.Predicates;

import net.minidev.json.JSONArray;
import net.minidev.json.parser.JSONParser;
import net.minidev.json.parser.ParseException;
import net.shibboleth.idp.attribute.AttributeDecodingException;
import net.shibboleth.idp.attribute.AttributeEncodingException;
import net.shibboleth.idp.attribute.IdPAttribute;
import net.shibboleth.idp.attribute.IdPAttributeValue;
import net.shibboleth.idp.attribute.StringAttributeValue;
import net.shibboleth.idp.attribute.transcoding.TranscodingRule;
import net.shibboleth.utilities.java.support.annotation.constraint.NonnullElements;
import net.shibboleth.utilities.java.support.annotation.constraint.NotEmpty;
import net.shibboleth.utilities.java.support.annotation.constraint.NotLive;
import net.shibboleth.utilities.java.support.annotation.constraint.Unmodifiable;

/**
 * Class encoding string attributes to string json object. Name of the attribute will be set as the key. The string
 * contains attribute value. If there are several attribute values they are delimited with delimiter(space is default)
 * or placed to array. If the value is set to be interpreted as int, only values that may be parsed to int are used. In
 * case output is interpreted as int but not set to array, only the first parsable value is used. If the value is
 * interpreted as boolean the value is true if string value equals to "true" ignoring the case. If boolean values are
 * not set to array the first string value is considered to be the result. Finally, the result may be placed to json
 * Object.
 */
public class OIDCStringAttributeTranscoder extends AbstractOIDCAttributeTranscoder {

    /** Whether to parse the values into and out as a JSON object. */
    @Nonnull @NotEmpty public static final String PROP_ASOBJECT = "oidc.asObject";

    /** Class logger. */
    @Nonnull private final Logger log = LoggerFactory.getLogger(OIDCStringAttributeTranscoder.class);

    /** {@inheritDoc} */
// Checkstyle: CyclomaticComplexity OFF
    @Override
    @Nullable protected Object encodeValues(@Nullable final ProfileRequestContext profileRequestContext,
            @Nonnull final IdPAttribute attribute, @Nonnull final TranscodingRule rule)
                    throws AttributeEncodingException {
        
        if (attribute.getValues().isEmpty()) {
            return null;
        }
        
        final List<?> values = transformValues(rule,
                attribute.getValues()
                    .stream()
                    .filter(StringAttributeValue.class::isInstance)
                    .map(StringAttributeValue.class::cast)
                    .map(StringAttributeValue::getValue));
        
        // Handle String values parsed into JSON.
        if (rule.getOrDefault(PROP_ASOBJECT, Boolean.class, false)) {
            if (values.size() == 1) {
                final Object value = values.get(0);
                if (!(value instanceof String)) {
                    throw new AttributeEncodingException("Value for IdPAttribute '" + attribute.getId() +
                            "' is defined to be parsed as JSON, but is not String-typed");
                }
                return toJSONObject((String) value);
            } else {
                final JSONArray array = new JSONArray();
                for (final Object value : values) {
                    if (!(value instanceof String)) {
                        log.warn("Value for IdPAttribute '{}' is defined to be parsed as JSON, but is not String-typed",
                                attribute.getId());
                        continue;
                    }
                    try {
                        array.add(toJSONObject((String) value));
                    } catch (final AttributeEncodingException e) {
                        log.warn("IdPAttribute '{}'", attribute.getId(), e);
                    }
                }
                return array;
            }
        }
        
        // Values of arbitrary mapped type placed into JSONArray.
        if (rule.getOrDefault(PROP_ASARRAY, Boolean.class, false)) {
            final JSONArray array = new JSONArray();
            array.addAll(values);
            return array;
        }
        
        // String concatenation / integer or boolean value.
        final String delimiter = rule.getOrDefault(PROP_STRING_DELIMITER, String.class, " ");
        
        boolean stringEmpty = true;
        final StringBuilder attributeString = new StringBuilder();
        for (final Object value : values) {
            if (value instanceof Integer || value instanceof Boolean) {
                log.warn("Values converted to Integer/Boolean but non-Array encoding, ignoring all but first value");
                return value;
            } else if (value instanceof String) {
                if (!stringEmpty) {
                    attributeString.append(delimiter);
                } else {
                    stringEmpty = false;
                }
                attributeString.append(value);
            } else {
                throw new AttributeEncodingException("Unrecognized value type: " + value.getClass().getName());
            }
        }
        
        return attributeString.length() != 0 ? attributeString.toString() : null;
    }
// Checkstyle: CyclomaticComplexity ON

    /** {@inheritDoc} */
    @Override
    @Nonnull @NonnullElements protected List<IdPAttributeValue> decodeValues(
            @Nullable final ProfileRequestContext profileRequestContext, @Nonnull final Object input,
            @Nonnull final TranscodingRule rule) throws AttributeDecodingException {
        
        // TODO: implement value decoding
        return Collections.emptyList();
    }
    
    /**
     * Parses string as JSONObject.
     * 
     * @param value string to parse
     * 
     * @return string parsed as JSONObject
     * 
     * @throws AttributeEncodingException if parsing fails 
     */
    @Nonnull private Object toJSONObject(@Nonnull @NotEmpty final String value) throws AttributeEncodingException {
        try {
            return new JSONParser(JSONParser.MODE_PERMISSIVE).parse(value);
        } catch (final ParseException e) {
            log.trace("Unable to parse string '{}' into JSONObject", value, e);
            throw new AttributeEncodingException("Unable to parse string into JSONObject", e);
        }
    }

    /**
     * Converts string values to List of Integer, String or Boolean depending on rule.
     * 
     * @param rule transcoding rule
     * @param attributeValues IdPAttribute values
     * 
     * @return list of String, Boolean or Integer objects
     */
    @Nonnull @NonnullElements @NotLive @Unmodifiable private List<?> transformValues(
            @Nonnull final TranscodingRule rule, @Nonnull @NonnullElements final Stream<String> attributeValues) {
        
        if (rule.getOrDefault(PROP_ASINTEGER, Boolean.class, false)) {
            log.debug("String values interpreted as integer by rule");
            return attributeValues
                    .map(v -> {
                        try {
                            return Integer.parseInt(v);
                        } catch (final NumberFormatException e) {
                            log.debug("Value '{}' not a parseable integer", v);
                            return null;
                        }
                    })
                    .filter(Predicates.notNull())
                    .collect(Collectors.toUnmodifiableList());
        }
        
        if (rule.getOrDefault(PROP_ASBOOLEAN, Boolean.class, false)) {
            log.debug("String values interpreted as Boolean by rule");
            return attributeValues
                    .map(v -> XSBooleanValue.valueOf(v).getValue())
                    .collect(Collectors.toUnmodifiableList());
        }

        return attributeValues.collect(Collectors.toUnmodifiableList());
    }

}