/*
 * Copyright (c) 2017 - 2020, GÉANT
 *
 * Licensed under the Apache License, Version 2.0 (the “License”); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an “AS IS” BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.geant.idpextension.oidc.profile.impl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.function.Function;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;

import net.minidev.json.JSONObject;
import net.shibboleth.idp.attribute.AttributeEncodingException;
import net.shibboleth.idp.attribute.IdPAttribute;
import net.shibboleth.idp.attribute.context.AttributeContext;
import net.shibboleth.idp.attribute.transcoding.AttributeTranscoder;
import net.shibboleth.idp.attribute.transcoding.AttributeTranscoderRegistry;
import net.shibboleth.idp.attribute.transcoding.TranscoderSupport;
import net.shibboleth.idp.attribute.transcoding.TranscodingRule;
import net.shibboleth.idp.profile.IdPEventIds;
import net.shibboleth.idp.profile.context.RelyingPartyContext;
import net.shibboleth.utilities.java.support.annotation.constraint.Live;
import net.shibboleth.utilities.java.support.annotation.constraint.NonnullAfterInit;
import net.shibboleth.utilities.java.support.annotation.constraint.NonnullElements;
import net.shibboleth.utilities.java.support.component.ComponentInitializationException;
import net.shibboleth.utilities.java.support.component.ComponentSupport;
import net.shibboleth.utilities.java.support.logic.Constraint;
import net.shibboleth.utilities.java.support.service.ReloadableService;
import net.shibboleth.utilities.java.support.service.ServiceableComponent;

import org.geant.idpextension.oidc.attribute.transcoding.OIDCAttributeTranscoder;
import org.geant.idpextension.oidc.messaging.context.OIDCAuthenticationResponseConsentContext;
import org.geant.idpextension.oidc.messaging.context.OIDCAuthenticationResponseTokenClaimsContext;
import org.geant.idpextension.oidc.profile.context.navigate.DefaultResponseClaimsSetLookupFunction;
import org.geant.idpextension.oidc.profile.context.navigate.OIDCAuthenticationResponseContextLookupFunction;
import org.opensaml.messaging.context.navigate.ChildContextLookup;
import org.opensaml.profile.action.ActionSupport;
import org.opensaml.profile.action.EventIds;
import org.opensaml.profile.context.ProfileRequestContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.nimbusds.oauth2.sdk.ResponseType;
import com.nimbusds.openid.connect.sdk.AuthenticationRequest;
import com.nimbusds.openid.connect.sdk.claims.ClaimsSet;

/**
 * Action that adds claims to a {@link ClaimsSet}. Claims are formed of resolved attributes having OIDC encoder. Action
 * verifies user has consented to release attribute, if consent information is available. Actions will not add claims
 * listed as reserved.
 */
public class AddAttributesToClaimsSet extends AbstractOIDCResponseAction {

    /** Class logger. */
    @Nonnull private Logger log = LoggerFactory.getLogger(AddAttributesToClaimsSet.class);

    /**
     * Strategy used to locate the {@link AttributeContext} associated with a given {@link ProfileRequestContext}.
     */
    @Nonnull private Function<ProfileRequestContext, AttributeContext> attributeContextLookupStrategy;

    /**
     * Strategy used to locate the response {@link ClaimsSet} associated with a given {@link ProfileRequestContext}.
     */
    @Nonnull private Function<ProfileRequestContext, ClaimsSet> responseClaimsSetLookupStrategy;

    /** Strategy used to locate the {@link OIDCAuthenticationResponseConsentContext}. */
    @Nonnull
    private Function<ProfileRequestContext, OIDCAuthenticationResponseConsentContext> consentContextLookupStrategy;
    
    /** Transcoder registry service object. */
    @NonnullAfterInit private ReloadableService<AttributeTranscoderRegistry> transcoderRegistry;

    /**
     * Whether attributes that result in an {@link net.shibboleth.idp.attribute.AttributeEncodingException}
     * when being encoded should be ignored or
     * result in an {@link net.shibboleth.idp.profile.IdPEventIds#UNABLE_ENCODE_ATTRIBUTE} transition.
     */
    private boolean ignoringUnencodableAttributes;
    
    /** AttributeContext to use. */
    @Nullable
    private AttributeContext attributeCtx;

    /** Claims Set to use. */
    @Nullable
    private ClaimsSet claimsSet;

    /** Whether we are adding claims to ID Token. */
    @Nonnull
    private boolean targetIDToken;

    /** Whether we can add claims to IDToken by default i.e. response type is "id_token". */
    @Nonnull
    private boolean addToIDTokenByDefault;

    /** List of claim names that will not be added. */
    @Nullable
    private List<String> reservedClaimNames;

    /** Constructor. */
    AddAttributesToClaimsSet() {
        attributeContextLookupStrategy = new ChildContextLookup<>(AttributeContext.class).compose(
                new ChildContextLookup<>(RelyingPartyContext.class));
        responseClaimsSetLookupStrategy = new DefaultResponseClaimsSetLookupFunction();
        consentContextLookupStrategy =
                new ChildContextLookup<>(OIDCAuthenticationResponseConsentContext.class).compose(
                        new OIDCAuthenticationResponseContextLookupFunction());
        
        ignoringUnencodableAttributes = true;
    }

    /**
     * Sets the registry of transcoding rules to apply to encode attributes.
     * 
     * @param registry registry service interface
     */
    public void setTranscoderRegistry(@Nonnull final ReloadableService<AttributeTranscoderRegistry> registry) {
        ComponentSupport.ifInitializedThrowUnmodifiabledComponentException(this);
        
        transcoderRegistry = Constraint.isNotNull(registry, "AttributeTranscoderRegistry cannot be null");
    }
    
    /**
     * Set whether the attributes that result in an {@link AttributeEncodingException} when being encoded
     * should be ignored or result in an {@link IdPEventIds#UNABLE_ENCODE_ATTRIBUTE} transition.
     * 
     * @param flag flag to set
     */
    public void setIgnoringUnencodableAttributes(final boolean flag) {
        ComponentSupport.ifInitializedThrowUnmodifiabledComponentException(this);

        ignoringUnencodableAttributes = flag;
    }
    
    /**
     * Set list of claim names that will not be added.
     * 
     * @param claimNames list of claim names that will not be added.
     */
    public void setReservedClaimNames(final List<String> claimNames) {
        ComponentSupport.ifInitializedThrowUnmodifiabledComponentException(this);
        reservedClaimNames = claimNames;
    }

    /**
     * Set whether target is id token claims set. If this flag is set addToIDTokenByDefault flag is active.
     * 
     * @param flag whether target is id token claims set
     */
    public void setTargetIDToken(final boolean flag) {
        ComponentSupport.ifInitializedThrowUnmodifiabledComponentException(this);
        targetIDToken = flag;
    }

    /**
     * Set the strategy used to locate the response {@link ClaimsSet} associated with a given
     * {@link ProfileRequestContext}.
     * 
     * @param strategy strategy used to locate the response {@link ClaimsSet} associated with a given
     *            {@link ProfileRequestContext}
     */
    public void setResponseClaimsSetLookupStrategy(@Nonnull final Function<ProfileRequestContext, ClaimsSet> strategy) {
        ComponentSupport.ifInitializedThrowUnmodifiabledComponentException(this);

        responseClaimsSetLookupStrategy =
                Constraint.isNotNull(strategy, "Response Claims Set lookup strategy cannot be null");
    }

    /**
     * Set the strategy used to locate the {@link AttributeContext} associated with a given
     * {@link ProfileRequestContext}.
     * 
     * @param strategy strategy used to locate the {@link AttributeContext} associated with a given
     *            {@link ProfileRequestContext}
     */
    public void setAttributeContextLookupStrategy(
            @Nonnull final Function<ProfileRequestContext, AttributeContext> strategy) {
        ComponentSupport.ifInitializedThrowUnmodifiabledComponentException(this);

        attributeContextLookupStrategy =
                Constraint.isNotNull(strategy, "AttributeContext lookup strategy cannot be null");
    }

    /**
     * Set the strategy used to locate the {@link OIDCAuthenticationResponseTokenClaimsContext} associated with a given
     * {@link ProfileRequestContext}.
     * 
     * @param strategy lookup strategy
     */
    public void setOIDCAuthenticationResponseConsentContextLookupStrategy(
            @Nonnull final Function<ProfileRequestContext, OIDCAuthenticationResponseConsentContext> strategy) {
        ComponentSupport.ifInitializedThrowUnmodifiabledComponentException(this);
        consentContextLookupStrategy = Constraint.isNotNull(strategy,
                "OIDCAuthenticationResponseConsentContext lookup strategy cannot be null");
    }

    /** {@inheritDoc} */
    @Override
    protected void doInitialize() throws ComponentInitializationException {
        super.doInitialize();
        
        if (transcoderRegistry == null) {
            throw new ComponentInitializationException("AttributeTranscoderRegistry cannot be null");
        }
    }
    
    /** {@inheritDoc} */
    @Override
    protected boolean doPreExecute(@Nonnull final ProfileRequestContext profileRequestContext) {
        if (!super.doPreExecute(profileRequestContext)) {
            return false;
        }
        
        attributeCtx = attributeContextLookupStrategy.apply(profileRequestContext);
        if (attributeCtx == null) {
            log.debug("{} No AttributeSubcontext available, nothing to do", getLogPrefix());
            return false;
        }
        
        claimsSet = responseClaimsSetLookupStrategy.apply(profileRequestContext);
        if (claimsSet == null) {
            log.error("{} No claims set to fill", getLogPrefix());
            ActionSupport.buildEvent(profileRequestContext, EventIds.INVALID_MSG_CTX);
            return false;
        }
        
        if (targetIDToken) {
            final Object msg = profileRequestContext.getInboundMessageContext().getMessage();
            if (msg instanceof AuthenticationRequest) {
                addToIDTokenByDefault =
                        !((AuthenticationRequest) msg).getResponseType().contains(ResponseType.Value.TOKEN);
            }
        }
        return true;
    }

    /** {@inheritDoc} */
    @Override
    protected void doExecute(@Nonnull final ProfileRequestContext profileRequestContext) {

        final List<JSONObject> claims = new ArrayList<>();
        
        ServiceableComponent<AttributeTranscoderRegistry> component = null;
        try {
            component = transcoderRegistry.getServiceableComponent();
            if (component == null) {
                log.error("Attribute transoding service unavailable");
                throw new AttributeEncodingException("Attribute transoding service unavailable");
            }
            
            for (final IdPAttribute attribute : attributeCtx.getIdPAttributes().values()) {
                if (attribute != null && !attribute.getValues().isEmpty()) {
                    encodeAttribute(component.getComponent(), profileRequestContext, attribute, claims);
                }
            }
        } catch (final AttributeEncodingException e) {
            ActionSupport.buildEvent(profileRequestContext, IdPEventIds.UNABLE_ENCODE_ATTRIBUTE);
            return;
        } finally {
            if (null != component) {
                component.unpinComponent();
            }
        }
        
        final OIDCAuthenticationResponseConsentContext consentCtx =
                consentContextLookupStrategy.apply(profileRequestContext);
        
        for (final JSONObject claim : claims) {
            for (final String name : claim.keySet()) {
                if (reservedClaimNames != null && reservedClaimNames.contains(name)) {
                    log.debug("{} claim has a reserved name {}. Not added to claims set", getLogPrefix(), name);
                    continue;
                }
                
                if (consentCtx != null && consentCtx.getConsentableAttributes().contains(name)
                        && !consentCtx.getConsentedAttributes().contains(name)) {
                    log.debug("{} Consentable attribute {} has no consent. Not added to claims set",
                            getLogPrefix(), name);
                    continue;
                }
                log.debug("{} Adding claim {} with value {}", getLogPrefix(), name, claim.get(name));
                claimsSet.setClaim(name, claim.get(name));
            }
        }
        
        log.debug("{} claims set after mapping attributes to claims {}", getLogPrefix(),
                claimsSet.toJSONObject().toJSONString());
    }

    /**
     * Access the registry of transcoding rules to transform the input attribute into claims.
     * 
     * @param registry  registry of transcoding rules
     * @param profileRequestContext current profile request context
     * @param attribute input attribute
     * @param results collection to add results to
     * 
     * @throws AttributeEncodingException if a non-ignorable error occurs
     */
    private void encodeAttribute(@Nonnull final AttributeTranscoderRegistry registry,
            @Nonnull final ProfileRequestContext profileRequestContext, @Nonnull final IdPAttribute attribute,
            @Nonnull @NonnullElements @Live final Collection<JSONObject> results)
                    throws AttributeEncodingException {
        
        final Collection<TranscodingRule> transcodingRules = registry.getTranscodingRules(attribute, JSONObject.class);
        if (transcodingRules.isEmpty()) {
            log.debug("{} Attribute {} does not have any transcoding rules, nothing to do", getLogPrefix(),
                    attribute.getId());
            return;
        }
        
        for (final TranscodingRule rule : transcodingRules) {
            try {
                if (targetIDToken) {
                    if (!addToIDTokenByDefault
                            && !rule.getOrDefault(OIDCAttributeTranscoder.PROP_FORCE_ID_TOKEN, Boolean.class, false)) {
                        log.debug("{} Attribute {} not targeted for ID Token", getLogPrefix(), attribute.getId());
                        continue;
                    }
                } else if (rule.getOrDefault(OIDCAttributeTranscoder.PROP_DENY_USERINFO, Boolean.class, false)) {
                    log.debug("{} Attribute {} not targeted for Userinfo response", getLogPrefix(), attribute.getId());
                    continue;
                }
                
                final AttributeTranscoder<JSONObject> transcoder = TranscoderSupport.<JSONObject>getTranscoder(rule);
                final JSONObject encodedAttribute =
                        transcoder.encode(profileRequestContext, attribute, JSONObject.class, rule);
                if (encodedAttribute != null) {
                    results.add(encodedAttribute);
                }
            } catch (final AttributeEncodingException e) {
                log.warn("{} Unable to encode attribute {}", getLogPrefix(), attribute.getId(), e);
                if (!ignoringUnencodableAttributes) {
                    throw e;
                }
            }
        }
    }

}