/*
 * Copyright (c) 2017 - 2020, GÉANT
 *
 * Licensed under the Apache License, Version 2.0 (the “License”); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an “AS IS” BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.geant.idpextension.oidc.profile.impl;

import java.time.Duration;
import java.time.Instant;
import java.util.function.Function;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;

import org.geant.idpextension.oidc.messaging.context.OIDCClientRegistrationResponseContext;
import org.geant.idpextension.oidc.metadata.resolver.ClientInformationManager;
import org.geant.idpextension.oidc.metadata.resolver.ClientInformationManagerException;

import org.opensaml.messaging.context.MessageContext;
import org.opensaml.messaging.context.navigate.ChildContextLookup;
import org.opensaml.profile.action.ActionSupport;
import org.opensaml.profile.action.EventIds;
import org.opensaml.profile.context.ProfileRequestContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.nimbusds.oauth2.sdk.client.ClientInformation;
import com.nimbusds.openid.connect.sdk.rp.OIDCClientInformation;
import com.nimbusds.openid.connect.sdk.rp.OIDCClientInformationResponse;

import net.shibboleth.idp.profile.AbstractProfileAction;
import net.shibboleth.utilities.java.support.annotation.constraint.NonNegative;
import net.shibboleth.utilities.java.support.annotation.constraint.NonnullAfterInit;
import net.shibboleth.utilities.java.support.component.ComponentInitializationException;
import net.shibboleth.utilities.java.support.component.ComponentSupport;
import net.shibboleth.utilities.java.support.logic.Constraint;

/**
 * An action that stores the {@link ClientInformation} from the {@link OIDCClientRegistrationResponseContext} to the
 * associated {@link ClientInformationManager}.
 */
public class StoreClientInformation extends AbstractProfileAction {

    /** Class logger. */
    @Nonnull private final Logger log = LoggerFactory.getLogger(StoreClientInformation.class);
    
    /** The client information manager used for storing the information. */
    @Nullable private ClientInformationManager clientInformationManager;
    
    /** Strategy to obtain registration validity period policy. */
    @Nullable private Function<ProfileRequestContext,Duration> registrationValidityPeriodStrategy;
    
    /** Default validity period for registration. */
    @Nullable @NonNegative private Duration defaultRegistrationValidityPeriod;
    
    /** The response message. */
    @Nullable private OIDCClientInformationResponse response;
    
    /**
     * Strategy used to locate the {@link OIDCClientRegistrationResponseContext} associated with a given 
     * {@link MessageContext}.
     */
    @Nonnull private Function<MessageContext,OIDCClientRegistrationResponseContext> oidcResponseContextLookupStrategy;

    /** Constructor. */
    public StoreClientInformation() {
        oidcResponseContextLookupStrategy = new ChildContextLookup<>(OIDCClientRegistrationResponseContext.class);
        defaultRegistrationValidityPeriod = Duration.ofHours(24);
    }
    
    /**
     * Set strategy function to obtain registration validity period.
     * 
     * @param strategy The strategy function.
     */
    public void setRegistrationValidityPeriodStrategy(
            @Nullable final Function<ProfileRequestContext,Duration> strategy) {
        ComponentSupport.ifInitializedThrowUnmodifiabledComponentException(this);
        
        registrationValidityPeriodStrategy = strategy;
    }
    
    /**
     * Set the default registration validity period in milliseconds.
     * 
     * @param lifetime The default validity period in milliseconds.
     */
    public void setDefaultRegistrationValidityPeriod(@NonNegative @Nonnull final Duration lifetime) {
        ComponentSupport.ifInitializedThrowUnmodifiabledComponentException(this);
        
        Constraint.isFalse(lifetime == null || lifetime.isNegative(),
                "Default registration validity period must be greater than or equal to 0");
        
        defaultRegistrationValidityPeriod = lifetime;
    }
    
    /**
     * Get the client information manager used for storing the information.
     * 
     * @return The client information manager used for storing the information
     */
    @NonnullAfterInit public ClientInformationManager getClientInformationManager() {
        return clientInformationManager;
    }
    
    /**
     * Set the client information manager used for storing the information.
     * @param manager The client information manager used for storing the information.
     */
    public void setClientInformationManager(@Nonnull final ClientInformationManager manager) {
        ComponentSupport.ifInitializedThrowUnmodifiabledComponentException(this);
        clientInformationManager = Constraint.isNotNull(manager, "The client information manager cannot be null!");
    }
    
    /**
     * Set the strategy used to locate the {@link OIDCClientRegistrationResponseContext} associated with a given
     * {@link MessageContext}.
     * 
     * @param strategy strategy used to locate the {@link OIDCClientRegistrationResponseContext} associated with a 
     *         given {@link MessageContext}
     */
    public void setOidcResponseContextLookupStrategy(
            @Nonnull final Function<MessageContext,OIDCClientRegistrationResponseContext> strategy) {
        ComponentSupport.ifInitializedThrowUnmodifiabledComponentException(this);
        
        oidcResponseContextLookupStrategy = Constraint.isNotNull(strategy,
                "OIDCClientRegistrationResponseContext lookup strategy cannot be null");
    }

    /** {@inheritDoc} */
    @Override
    protected void doInitialize() throws ComponentInitializationException {
        super.doInitialize();
        
        if (clientInformationManager == null) {
            throw new ComponentInitializationException("ClientInformationManager cannot be null");
        }
    }

    /** {@inheritDoc} */
    @Override
    protected boolean doPreExecute(@Nonnull final ProfileRequestContext profileRequestContext) {
        if (!super.doPreExecute(profileRequestContext)) {
            return false;
        }
        
        if (profileRequestContext.getOutboundMessageContext() == null) {
            log.error("{} Unable to locate outbound message context", getLogPrefix());
            ActionSupport.buildEvent(profileRequestContext, EventIds.INVALID_MSG_CTX);
            return false;
        }
        
        final Object message = profileRequestContext.getOutboundMessageContext().getMessage();
        if (message == null || !(message instanceof OIDCClientInformationResponse)) {
            log.error("{} Unable to locate outbound message", getLogPrefix());
            ActionSupport.buildEvent(profileRequestContext, EventIds.INVALID_MSG_CTX);
            return false;
        }
        response = (OIDCClientInformationResponse) message;
        return true;
    }
    
    /** {@inheritDoc} */
    @Override
    protected void doExecute(@Nonnull final ProfileRequestContext profileRequestContext) {

        final OIDCClientInformation clientInformation = response.getOIDCClientInformation();
        final Duration lifetime = registrationValidityPeriodStrategy != null ?
                registrationValidityPeriodStrategy.apply(profileRequestContext) : null;
        
        try {
            if (lifetime != null && lifetime.isZero()) {
                log.debug("{} Registration won't expire, lifetime set to 0", getLogPrefix());
                clientInformationManager.storeClientInformation(clientInformation, null);
            } else {
                if (lifetime == null) {
                    log.debug("{} No registration validity period supplied, using default", getLogPrefix());
                }
                final Instant expiration = Instant.now().plus(
                        lifetime != null ? lifetime : defaultRegistrationValidityPeriod);
                log.debug("{} Registration will expire on {}", getLogPrefix(), expiration);
                clientInformationManager.storeClientInformation(clientInformation, expiration);                
            }
        } catch (final ClientInformationManagerException e) {
            log.error("{} Could not store the client information", getLogPrefix(), e);
            ActionSupport.buildEvent(profileRequestContext, EventIds.IO_ERROR);
            return;
        }
        log.info("{} Client information successfully stored for {}", getLogPrefix(), 
                clientInformation.getID().getValue());
    }
    
}